///////////////////////////////////////////////////////////////////////
//
// This file is part of the finlink conduit for PalmCoins 
// Copyright (C) 2001 Andrew S. Townley 
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA,
// 02111-1307  USA
//
// Name:	GNCAccountManager.cpp
// Author:	Andrew S. Townley <adz1092 at netscape dot net>
// Created:	Sat May 26 11:42:41 CDT 2001
//
///////////////////////////////////////////////////////////////////////

#include "GNCAccountManager.h"
#include "ManagerCreator.h"

#define MATURITY_LEVEL	0

using namespace std;

/**
 * This is a private utility function that checks the status of the
 * GNUCash session structure.
 */

static void session_check(Session *sess)
{
	if(sess == static_cast<Session*>(NULL))
		return;

	int err = xaccSessionGetError(sess);
	if(err != 0)
		{
		throw FatalException("error:  %s", strerror(err));
		}
}

/**
 * Utility function to do the recursive loading of the AccountNode
 * information from the GNUCash account information.
 *
 * @param node the parent node
 * @param g the corresponding parent group
 */

static void read_helper(GNCAccountNode *node, AccountGroup *g)
{
	if(node == static_cast<GNCAccountNode*>(NULL) ||
			g == static_cast<AccountGroup*>(NULL))
		{
		return;
		}

	int count = xaccGroupGetNumAccounts(g);
	for(int i = 0; i < count; ++i)
		{
		// process the account
		Account *a = xaccGroupGetAccount(g, i);

//cerr << "Read account:  " << xaccAccountGetName(a) << endl;

		AccountGroup *akidz = xaccAccountGetChildren(a);
		GNCAccountNode *an = new GNCAccountNode(a);
		node->add(an, true);

//cerr << "Added to:  " << node->getName() << endl;

		// process the account children
		read_helper(an, akidz);
		}
}

GNCAccountManager::GNCAccountManager(const string& uri)
	throw(FatalException)
{
	static const char *method = "GNCAccountManager";
	trace.init(method, this, MATURITY_LEVEL);
	trace.methodStart(method);
	trace.methodArg("uri", uri);

	xaccConfigSetForceDoubleEntry(0);
	sess = xaccMallocSession();
	if(sess == static_cast<Session*>(NULL))
		{
		trace.methodThrow(method, "FatalException");
		throw FatalException("error:  out of memory (session allocation failed)");
		}

	if(uri.length())
		{
		// open the session
		xaccSessionBegin(sess, uri.c_str());
		session_check(sess);
		group = xaccSessionGetGroup(sess);
		session_check(sess);

		root = new GNCAccountNode();
		}

	trace.methodReturn(method);
}

GNCAccountManager::~GNCAccountManager()
{
	static const char *method = "~GNCAccountManager";
	trace.methodStart(method);

	// shut down the session
	if(sess != static_cast<Session*>(NULL))
		{
		xaccSessionEnd(sess);
		xaccSessionDestroy(sess);
		xaccFreeAccountGroup(group);
		}
	
	// delete our parallel root account
	delete root;
	root = static_cast<GNCAccountNode*>(NULL);

	// delete the transactions
	TransList::iterator p;
	for(p = _transactions.begin(); p != _transactions.end(); ++p)
		{
		delete *p;
		}
	
	trace.methodReturn(method);
}

AccountNode::iterator* GNCAccountManager::accounts()
{
	static const char *method = "accounts";
	trace.methodStart(method);
	AccountNode::iterator *x = root->children();

	trace.methodReturn(method, static_cast<void*>(x));
	return root->children();
}

void GNCAccountManager::addAccount(AccountNode *a)
{
	static const char *method = "addAccount";
	trace.methodStart(method);
	trace.methodArg("a", a);

	root->add(a);
	
	trace.methodReturn(method);
}

void GNCAccountManager::deleteAccount(AccountNode *a)
{
	static const char *method = "deleteAccount";
	trace.methodStart(method);
	trace.methodArg("a", a);
	
	root->remove(a);
	trace.methodReturn(method);
}

void GNCAccountManager::deleteAllAccounts()
{
	static const char *method = "deleteAllAccounts";
	trace.methodStart(method);

	AccountNode::iterator_ptr p(root->children());
	for(p->first(); p->moreElements(); p->next())
		{
		AccountNode *n = p->current();
		root->remove(n);

		// clean up the other stuff
		delete n;
		}
	
	trace.methodReturn(method);
}

string GNCAccountManager::getUsage()
{
	static const char *method = "getUsage";
	static const char *s = "file:/home/somewhere/somedir/file.xac";

	trace.methodStart(method);
	trace.methodReturn(method, s);
	return string(s);
}

AccountNode* GNCAccountManager::newNode()
{
	static const char *method = "newNode";

	trace.methodStart(method);
	AccountNode *p = root->newInstance();

	trace.methodReturn(method, static_cast<void*>(p));
	return p;
}

void GNCAccountManager::readAccounts() throw(FatalException)
{
	static const char *method = "readAccounts";

	trace.methodStart(method);
	try
		{
		// recursively process all the accounts
		read_helper(root, group);
		}
	catch(const FatalException& e)
		{
		trace.methodThrow(method, "FatalException");
		throw;
		}
	trace.methodReturn(method);
}

void GNCAccountManager::writeAccounts() throw(FatalException)
{
	static const char *method = "writeAccounts";

	trace.methodStart(method);
	if(sess != static_cast<Session*>(NULL))
		{
		xaccSessionSetGroup(sess, root->getChildAccountGroup());
		xaccSessionSave(sess);
		}
	trace.methodReturn(method);
}

void GNCAccountManager::addTransaction(TransactionNode *n)
{
	static const char *method = "addTransaction";

	trace.methodStart(method);
	trace.methodArg("n", n);
	_transactions.push_back(n);
	trace.methodReturn(method);
}

class TransFind
{
public:
	TransFind(TransactionNode *n) : _targ(n) {}
	bool operator()(TransactionNode *x) const
	{
		return *_targ == *x;
	}

protected:
	TransactionNode *_targ;
};

TransactionNode* GNCAccountManager::getTransaction(TransactionNode *n)
{
	static const char *method = "getTransaction";

	trace.methodStart(method);
	trace.methodArg("n", n);

	TransList::iterator p = find_if(_transactions.begin(),
				_transactions.end(), TransFind(n));

	TransactionNode *tn = static_cast<TransactionNode*>(NULL);
	if(p != _transactions.end())
		{
		tn = *p;
		}
	
	trace.methodReturn(method, tn);
	return tn;
}
