///////////////////////////////////////////////////////////////////////
//
// This file is part of the finlink conduit for PalmCoins 
// Copyright (C) 2001 Andrew S. Townley 
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA,
// 02111-1307  USA
//
// Name:	GNCAccountNode.cpp
// Author:	Andrew S. Townley <adz1092 at netscape dot net>
// Created:	Sat May 26 11:42:41 CDT 2001
//
///////////////////////////////////////////////////////////////////////

#include "GNCAccountNode.h"
#include "GNCTransactionNode.h"
#include "GNCTransSplit.h"

#define MATURITY_LEVEL 0

using namespace std;

template <class T>
class GNCTransIterator : public Iterator<T>
{
public:
	GNCTransIterator(Account *t)
		: _acc(t), _split(static_cast<Split*>(NULL)),
		  _trans(static_cast<Transaction*>(NULL)), 
		  _gt(static_cast<GNCTransactionNode*>(NULL)), _idx(0)
	{
	}

	void help()
	{
		Transaction *tt;

		_split = xaccAccountGetSplit(_acc, _idx);
		tt = xaccSplitGetParent(_split);
		if(tt != _trans)
			{
			_trans = tt;
			}
		if(_gt != static_cast<GNCTransactionNode*>(NULL))
			{
			delete _gt;
			}
		_gt = new GNCTransactionNode(_trans, _acc);
	}

	void first()
	{
		_idx = 0;
		help();
	}

	void next()
{
		_idx++;
		help();
	}

	bool moreElements() const
	{
		if(_idx < xaccAccountGetNumSplits(_acc))
			return true;

		return false;
	}

	T current() const
	{
		return _gt;
	}

	~GNCTransIterator()
	{
		if(psplit != static_cast<TransSplit*>(NULL))
			{
			delete psplit;
			}
	}

protected:
	Account 		*_acc;
	Split			*_split;
	Transaction		*_trans;
	GNCTransactionNode	*_gt;
	int			_idx;
};

template <class T>
class GNCAccSplitIt : public Iterator<T>
{
public:
	GNCAccSplitIt(Account *t)
		: _acc(t), _split(static_cast<GNCTransSplit*>(NULL)),
		  _idx(0)
	{
	}

	void help()
	{
		Split *s = xaccAccountGetSplit(_acc, _idx);
		if(_split != static_cast<GNCTransSplit*>(NULL))
			{
			delete _split;
			}
		_split = new GNCTransSplit(s);
	}

	void first()
	{
		_idx = 0;
		help();
	}

	void next()
	{
		_idx++;
		help();
	}

	bool moreElements() const
	{
		if(_idx < xaccAccountGetNumSplits(_acc))
			return true;

		return false;
	}

	T current() const
	{
		return _split;
	}

	~GNCAccSplitIt()
	{
		if(_gt != static_cast<GNCTransSplit*>(NULL))
			{
			delete _gt;
			}
	}

protected:
	Account 		*_acc;
	GNCTransSplit		*_split;
	int			_idx;
};

/**
 * This is the default constructor.  Since we must maintain the proper
 * references between the account nodes and the corresponding GNUCash
 * accounts, we have to allocate a GNUCash Account.
 */

GNCAccountNode::GNCAccountNode()
	: acc(static_cast<Account*>(NULL)),
	  _manage_account(true),
	  _p(static_cast<GNCAccountNode*>(NULL))
{
	static const char *m = "GNCAccountNode";
	trace.init(m, this, MATURITY_LEVEL);
	trace.methodStart(m);

	// allocate the underlying GNC account node
	acc = xaccMallocAccount();
	xaccInitAccount(acc);

	trace.methodReturn(m);
}

/**
 * This constructor just needs to cache the reference to this
 * particular structure.  We'll always be manipulating it directly.
 *
 * @param acc the GNUCash Account structure
 */

GNCAccountNode::GNCAccountNode(Account *a)
	: acc(a), _manage_account(false),
	  _p(static_cast<AccountNode*>(NULL))
{
	static const char *m = "GNCAccountNode";
	trace.init(m, this, MATURITY_LEVEL);
	trace.methodStart(m);
	trace.methodArg("a", a);

	trace.methodReturn(m);
}

/**
 * The destructor is responsible for freeing the child accounts.
 */

GNCAccountNode::~GNCAccountNode()
{
	static const char *m = "~GNCAccountNode";
	trace.methodStart(m);

	// clean up the GNC data
	if(acc != static_cast<Account*>(NULL) && _manage_account)
		{
//		xaccFreeAccount(acc);
		acc = static_cast<Account*>(NULL);
		}

	ChildList::iterator p;
	for(p = _children.begin(); p != _children.end(); ++p)
		{
		delete *p;
		}
	_children.erase(_children.begin(), _children.end());

	trace.methodReturn(m);
}
	
//GUID GNCAccountNode::getId() const
//{
//	return xaccAccountGetGUID(acc);
//}

AccountGroup* GNCAccountNode::getChildAccountGroup()
{
	static const char *m = "getChildAccountGroup";
	trace.methodStart(m);

	AccountGroup *g = xaccAccountGetChildren(acc);
	trace.methodReturn(m, g);
	return g;
}

string GNCAccountNode::getName()
{
	static const char *m = "getName";
	trace.methodStart(m);
	string s;

	if(acc != static_cast<Account*>(NULL))
		{
		s = xaccAccountGetName(acc);
		}

	trace.methodReturn(m, s);
	return s;
}

void GNCAccountNode::setName(const string& s)
{
	static const char *m = "setName";
	trace.methodStart(m);
	trace.methodArg("s", s);
	
	if(acc != static_cast<Account*>(NULL))
		{
		xaccAccountSetName(acc, s.c_str());
		}
	
	trace.methodReturn(m);
}

string GNCAccountNode::getDescription()
{
	static const char *m = "getDescription";
	trace.methodStart(m);
	string s;
	
	if(acc != static_cast<Account*>(NULL))
		{
		s = xaccAccountGetDescription(acc);
		}
	
	trace.methodReturn(m, s);
	return s;
}

void GNCAccountNode::setDescription(const string& s)
{
	static const char *m = "setDescription";
	trace.methodStart(m);
	trace.methodArg("s", s);

	if(acc != static_cast<Account*>(NULL))
		{
		xaccAccountSetDescription(acc, s.c_str());
		}
	
	trace.methodReturn(m);
}

double GNCAccountNode::getBalance()
{
	static const char *m = "getBalance";
	trace.methodStart(m);
	
	double d = 0.0;
	
	if(acc != static_cast<Account*>(NULL))
		{
		d = xaccAccountGetBalance(acc);
		}

	trace.methodReturn(m, d);
	return d;
}

void GNCAccountNode::setBalance(double d)
{
	static const char *m = "setBalace";
	trace.methodStart(m);
	trace.methodArg("d", d);

	if(acc != static_cast<Account*>(NULL))
		{
		// FIXME: shouldn't manipulate this directly!
		acc->balance = d;
		}
	
	trace.methodReturn(m);
}

AccountNode::AccountType GNCAccountNode::getType()
{
	static const char *m = "getType";
	trace.methodStart(m);

	AccountNode::AccountType t;
	GNCAccountType at;
	
	if(acc != static_cast<Account*>(NULL))
		{
		at = xaccAccountGetType(acc);
		switch(at)
			{
			case ::BANK:
				t = AccountNode::BANK;
				break;
			case ::CASH:
				t = AccountNode::CASH;
				break;
			case ::INCOME:
				t = AccountNode::INCOME;
				break;
			case ::CREDIT:
				t = AccountNode::CREDIT;
				break;
			case ::EXPENSE:
				t = AccountNode::EXPENSE;
				break;
			default:
				t = AccountNode::INVALID;
				break;
			}
		}

	trace.methodReturn(m, t);
	return t;
}

void GNCAccountNode::setType(AccountType t)
{
	static const char *m = "setType";
	trace.methodStart(m);
	trace.methodArg("t", t);

	if(acc == static_cast<Account*>(NULL))
		{
		trace.tprintf(5, "account was null");
		trace.methodReturn(m);
		return;
		}

	switch(t)
		{
		case AccountNode::BANK:
			xaccAccountSetType(acc, ::BANK);
			break;
		case AccountNode::CASH:
			xaccAccountSetType(acc, ::CASH);
			break;
		case AccountNode::INCOME:
			xaccAccountSetType(acc, ::INCOME);
			break;
		case AccountNode::CREDIT:
			xaccAccountSetType(acc, ::CREDIT);
			break;
		case AccountNode::EXPENSE:
			xaccAccountSetType(acc, ::EXPENSE);
			break;
		}
	
	trace.methodReturn(m);
}

AccountNode* GNCAccountNode::getParent()
{
	static const char *m = "getParent";
	trace.methodStart(m);
	trace.methodReturn(m, _p);
	
	return _p;
}

void GNCAccountNode::setParent(AccountNode *p)
{
	static const char *m = "setParent";
	trace.methodStart(m);
	trace.methodArg("p", p);

	_p = p;

	trace.methodReturn(m);
}

void GNCAccountNode::add(AccountNode *node, bool b)
{
	static const char *m = "add";
	trace.methodStart(m);
	trace.methodArg("node", node);
	trace.methodArg("b", b);

	node->setParent(this);
	_children.push_back(node);

	trace.methodReturn(m);
}

void GNCAccountNode::add(AccountNode *node)
{
	static const char *m = "add";
	trace.methodStart(m);
	trace.methodArg("node", node);

	node->setParent(this);
	_children.push_back(node);

	GNCAccountNode *gnc;
	if(gnc = (GNCAccountNode*)node)
		{
		xaccRemoveAccount(gnc->acc);
		xaccInsertSubAccount(acc, gnc->acc);
//cout << "acc = " << acc << "; gnc->acc parent:  " << xaccAccountGetParentAccount(gnc->acc) << endl;
		}
	
	trace.methodReturn(m);
}

void GNCAccountNode::remove(AccountNode *node)
{
	static const char *m = "remove";
	trace.methodStart(m);
	trace.methodArg("node", node);

	ChildList::iterator p = find(_children.begin(), 
			_children.end(), node);
	if(p != _children.end())
		{
		// remove the GNUCash structure
		GNCAccountNode *gnc;
		if(gnc = static_cast<GNCAccountNode*>(node))
			{
			xaccRemoveAccount(gnc->acc);
			}

		_children.erase(p);
		node->setParent(static_cast<AccountNode*>(NULL));
		}
	
	trace.methodReturn(m);
}

AccountNode::iterator* GNCAccountNode::children()
{
	static const char *m = "children";
	trace.methodStart(m);

	AccountNode::iterator *i = new ListIterator<AccountNode*>(_children);
	
	trace.methodReturn(m, i);
	return i;
}

TransactionNode* GNCAccountNode::newTransaction()
{
	static const char *m = "newTransaction";
	trace.methodStart(m);
	TransactionNode *t = new GNCTransactionNode();
	trace.methodReturn(m, t);
	return t;
}

void GNCAccountNode::deleteTransaction(TransactionNode *n)
{
	static const char *m = "deleteTransaction";
	trace.methodStart(m);
	trace.methodArg("n", n);

	trace.tprintf(5, "method not implemented");

	trace.methodReturn(m);
}

void GNCAccountNode::deleteAllTransactions()
{
	static const char *m = "deleteAllTransactions";
	trace.methodStart(m);

	trace.tprintf(5, "method not implemented");

	trace.methodReturn(m);
}

TransactionNode::iterator* GNCAccountNode::transactions()
{
	static const char *m = "transactions";
	trace.methodStart(m);

	TransactionNode::iterator *i = new GNCTransIterator<TransactionNode*>(acc);

	trace.methodReturn(m, i);
	return i;
}

TransSplit* GNCAccountNode::newSplit()
{
	static const char *m = "newSplit";
	trace.methodStart(m);

	TransSplit *s = new GNCTransSplit();
	trace.methodReturn(m, s);
	return s;
}

static void dump_splits(ErrorTrace& trace, Account *a)
{
	Transaction *last_t = 0;

	int c = xaccAccountGetNumSplits(a);
	for(int i = 0; i < c; ++i)
		{
		Split* s = xaccAccountGetSplit(a, i);
		Transaction *t = xaccSplitGetParent(s);
		if(t != last_t)
			{
			trace.tprintf(5, Format::make_line('-', 80).c_str());
				trace.tprintf(5, "%10s|%-6s|%-38s",
					Format::fmt_date(xaccTransGetDate(t)).c_str(),
					xaccTransGetNum(t),
					Format::elid_string(xaccTransGetDescription(t), 38).c_str());
			last_t = t;

			int c2 = xaccTransCountSplits(t);
			for(int j = 0; j < c2; ++j)
				{
				Split *ts = xaccTransGetSplit(t, j);
				const char *name = xaccAccountGetName(xaccSplitGetAccount(ts));
				const char *memo = xaccSplitGetMemo(ts);
				trace.tprintf(5, "%10d*%-38s*%18s*%10.2lf",
					j, Format::elid_string((memo ? memo : ""), 38).c_str(),
					Format::elid_string((name ? name : ""), 18).c_str(),
					xaccSplitGetValue(ts));
				}
			}
		trace.tprintf(5, "%10d|%-6s|%-38s|%10.2lf|%10.2lf",
			i, xaccSplitGetAction(s), 
			Format::elid_string(xaccSplitGetMemo(s), 38).c_str(),
			xaccSplitGetValue(s), xaccSplitGetBalance(s));
		}
	trace.tprintf(5, Format::make_line('-', 80).c_str());
}

void GNCAccountNode::addSplit(TransSplit *s)
{
	static const char *m = "addSplit";
	trace.methodStart(m);
	trace.methodArg("s", s);

	if(trace.willTrace(5))
		{
		trace.tprintf(5, "Balance for account '%s' before new split:  %10.2lf", getName().c_str(), xaccAccountGetBalance(acc));
		dump_splits(trace, acc);
		}

	GNCTransSplit* gs;
	if(gs = static_cast<GNCTransSplit*>(s))
		{
		xaccAccountInsertSplit(acc, gs->getGNCSplit());
		gs->_manage = false;

		if(trace.willTrace(5))
			{
			Split* _split = gs->getGNCSplit();
			trace.tprintf(5, "Transaction: %s", xaccTransGetDescription(xaccSplitGetParent(_split)));
			trace.tprintf(5, "\tAccount:   %s", xaccAccountGetName(xaccSplitGetAccount(_split)));
			trace.tprintf(5, "\tmemo:      %s", xaccSplitGetMemo(_split));
			trace.tprintf(5, "\tvalue:     %10.2lf", xaccSplitGetValue(_split));
			trace.tprintf(5, "\tbalance:   %10.2lf", xaccSplitGetBalance(_split));
			trace.tprintf(5, "Balance after new split:  %10.2lf", xaccAccountGetBalance(acc));
			dump_splits(trace, acc);
			}
		}
	
	trace.methodReturn(m);
}

void GNCAccountNode::deleteSplit(TransSplit *s)
{
	static const char *m = "addSplit";
	trace.methodStart(m);
	trace.methodArg("s", s);
	
	GNCTransSplit *gs;
	if(gs = static_cast<GNCTransSplit*>(s))
		{
		xaccSplitDestroy(gs->getGNCSplit());
		}
	
	trace.methodReturn(m);
}

TransSplit::iterator* GNCAccountNode::splits()
{
	static const char *m = "splits";
	trace.methodStart(m);

	TransSplit::iterator* i = new GNCAccSplitIt<TransSplit*>(acc);

	trace.methodReturn(m, i);
	return i;
}

string GNCAccountNode::toString()
{
	return getPath("::");
}
