#ifndef ACCOUNTNODE_H
#define ACCOUNTNODE_H
///////////////////////////////////////////////////////////////////////
//
// This file is part of the finlink conduit for PalmCoins 
// Copyright (C) 2001 Andrew S. Townley 
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA,
// 02111-1307  USA
//
// Name:	AccountNode.h
// Author:	Andrew S. Townley <adz1092 at netscape dot net>
// Created:	Sat Jun  2 13:16:31 CDT 2001
//
///////////////////////////////////////////////////////////////////////

#include "Iterator.h"
#include "TransactionNode.h"

/**
 * This class takes care of the implementation of the forward iterator
 * operatons for the STL list.
 */

template <class T>
class ListIterator : public Iterator<T>
{
	typedef std::list<T> Sequence;

public:
	ListIterator(const Sequence& cl)
		: started(false), l(cl)
	{
	}

	virtual void first()
	{
		p = l.begin();
		started = true;
	}
		
	virtual void next()
	{
		if(!started)
			{
			first();
			}
		else
			p++;
	}

	virtual bool moreElements() const
	{
		if(p != l.end())
			return true;

		return false;
	}

	virtual T current() const
	{
		return *p;
	}

protected:
	bool				started;
	Sequence			l;
	Sequence::const_iterator	p;
};

/**
 * This class represents the AccountNode product that is created by
 * the AccountNodeBuilder to represent a generic financial account.
 * We have to define common functionality (LCD) so that the accounts
 * can be adequately compared.
 * <p>
 * In both GNUCash and PalmCoins, accounts can be nested.  Since the
 * nesting can be arbitrary in both, we're defining a single,
 * composite class to handle the details of the account management.
 * There will always be a root account node that shouldn't be
 * displayed or used in any comparison operations.
 * </p>
 * <p>
 * Memory management of the children in the compisite is handled by
 * the parent, so if the parent is deleted, all references to child
 * nodes become invalid.
 * </p>
 * <p>
 * In order for proper data integrity to be maintained, each product
 * that is participating in the sync operation must have a
 * product-specific account class derived from AccountNode, otherwise
 * data could be lost.
 * </p>
 */

class AccountNode : public Traceable
{
public:
	/**
	 * The AccountType definition includes a union of the
	 * supported account types from all the packages that we can
	 * sync with.  Currently, it borrows some types from GNUCash.
	 */

	enum AccountType
	{
		INVALID = -1,

		/** a regular bank account */
		BANK = 0,

		/** a cash repository of hard currency */
		CASH,

		/**
		 * a credit account such as a credit card or other
		 * line of credit
		 */

		CREDIT,

		/** generic income source */
		INCOME,

		/** generic expense */
		EXPENSE
	};

	typedef Iterator<AccountNode*>		iterator;
	typedef IteratorPtr<AccountNode*>	iterator_ptr;

	virtual std::string getName() { return ""; }
	virtual void setName(const std::string& s) {}
	virtual std::string getDescription() { return ""; }
	virtual void setDescription(const std::string& s) {}
	virtual double getBalance() { return 0.0; }
	virtual AccountType getType() { return INVALID; }
	virtual void setType(AccountType t) {}

	/**
	 * This method is used to return a pointer to the parent node.
	 * The top of the tree should return NULL indicating that
	 * there are no more parents.
	 *
	 * @return the parent node or NULL if the root of the tree
	 */

	virtual AccountNode* getParent() { return 0; }
	virtual void setParent(AccountNode *p) {}

	/**
	 * This method adds an account as a child to this account.
	 *
	 * @param node the account to add
	 */

	virtual void add(AccountNode *node) {}

	/**
	 * This method will remove the indicated account as a child of
	 * the current account.  Once the node has been removed,
	 * memory management of the account that was removed is the
	 * responsibility of the caller.
	 *
	 * @param node the account to remove
	 */

	virtual void remove(AccountNode *node) {}

	/**
	 * This method returns way to traverse the child accounts.
	 * NOTE:  The iterator must be wrapped with an IteratorPtr
	 * instance in order to prevent memory leaks.
	 * <pre>
	 * IteratorPtr<AccountNode*> it(node->children());
	 * </pre>
	 *
	 * @return a foward iterator for any child accounts
	 */

	virtual iterator* children() { return NULL; }

	/**
	 * This method returns a path representation of the specified
	 * account, based on the path separator string.
	 *
	 * @param sep the path separator
	 * @return the fully-qualified path of the account node
	 */

	virtual std::string getPath(const std::string& sep)
	{
		std::string s;

		AccountNode *p = static_cast<AccountNode*>(NULL);
		for(p = this; p != static_cast<AccountNode*>(NULL);
				p = p->getParent())
			{
			std::string t = p->getName();
			if(t.length())
				{
				s = sep + t + s;
				}
			}

		return s;
	}
	
	/**
	 * This method is used to create a new instance of the
	 * specified sub-class.  All attributes are initialized to
	 * their default values.  It is *NOT* a clone() operation.
	 *
	 * @return a new instance of the derivied class
	 */

	virtual AccountNode *newInstance() = 0;

	/**
	 * This method performs a deep copy of the source AccountNode
	 * into the destination AccountNode.
	 *
	 * @param src the source AccountNode
	 */

	virtual void copy(AccountNode *src)
	{
		setName(src->getName());
		setDescription(src->getDescription());
//		setBalance(src->getBalance());
		setType(src->getType());

		iterator_ptr p(src->children());
		for(p->first(); p->moreElements(); p->next())
			{
			AccountNode *newChild = newInstance();
			add(newChild);
			newChild->copy(p->current());
			}
	}

	virtual ~AccountNode() {};

	// support for transactions
	virtual TransactionNode* newTransaction() = 0;
	virtual void deleteTransaction(TransactionNode *t) = 0;
	virtual void deleteAllTransactions() = 0;
	virtual TransactionNode::iterator* transactions() = 0;

	// support for Splits
	virtual TransSplit* newSplit() = 0;
	virtual void addSplit(TransSplit *s) = 0;
	virtual void deleteSplit(TransSplit *s) = 0;
	virtual TransSplit::iterator* splits() = 0;

protected:
	AccountNode() {};
	virtual void setBalance(double d) {};
};

#endif
