#ifndef ITERATOR_H
#define ITERATOR_H
///////////////////////////////////////////////////////////////////////
//
// This file is part of the finlink conduit for PalmCoins 
// Copyright (C) 2001 Andrew S. Townley 
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA,
// 02111-1307  USA
//
// Name:	Iterator.h
// Author:	Andrew S. Townley <adz1092 at netscape dot net>
// Created:	Sat Jun  2 13:08:49 CDT 2001
//
///////////////////////////////////////////////////////////////////////

#include "stdcxx.h"

/**
 * This class represents a generic iterator.  We are borrowing it from
 * the design pattern book, but the actual implementation may be in
 * terms of the Standard C++ Library's iterator class.
 */

template <class T>
class Iterator
{
public:
	/**
	 * This method positions the iterator to the first element of
	 * the sequence.
	 */

	virtual void first() = 0;

	/**
	 * This method advances the iterator to the next element in
	 * the sequence.  If there is not a current element, calling
	 * this method is equivalent to calling the first() method.
	 */

	virtual void next() = 0;

	/**
	 * This method is used to determine if there are more elements
	 * in the sequence.
	 *
	 * @return true if there are more elements; false otherwise
	 */

	virtual bool moreElements() const = 0;

	/**
	 * This method returns an object in the sequence from the
	 * current position of the iterator.
	 *
	 * @return an object of the specified type
	 */

	virtual T current() const = 0;

protected:
	Iterator() {}
};

/**
 * This is a utility class to help manage the memory of the Iterator
 * pointers returned by the methods that return iterators.  The proper
 * use of this class is:
 * <pre>
 * IteratorPtr<int>	p(somecls->getInts());
 * </pre>
 * where the getInts() method returns an Iterator<int>*.
 */

template <class T>
class IteratorPtr
{
public:
	IteratorPtr(Iterator<T>* i) : _i(i) {}
	~IteratorPtr() { delete _i; }
	Iterator<T>* operator->() { return _i; }
	Iterator<T>& operator*() { return *_i; }

private:
	IteratorPtr(const IteratorPtr&);
	IteratorPtr& operator=(const IteratorPtr&);

private:
	Iterator<T>* _i;
};

#endif
