#ifndef MANAGERFACTORY_H
#define MANAGERFACTORY_H
///////////////////////////////////////////////////////////////////////
//
// This file is part of the finlink conduit for PalmCoins 
// Copyright (C) 2001 Andrew S. Townley 
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA,
// 02111-1307  USA
//
// Name:	ManagerFactory.h
// Author:	Andrew S. Townley <adz1092 at netscape dot net>
// Created:	Sat Jun  2 19:29:20 CDT 2001
//
///////////////////////////////////////////////////////////////////////

#include "stdcxx.h"
#include "AccountManager.h"

class ManagerCreator;

/**
 * This class exists as a singleton object to provide a registration
 * database for available AccountManager implementations.  Each
 * implementation should register themselves with the manager factory
 * in the library where they are implemented.
 */

class ManagerFactory
{
public:
	typedef Iterator<std::string>		iterator;
	typedef IteratorPtr<std::string>	iterator_ptr;

public:
	/**
	 * This static method assures that we only ever have a single
	 * instance of the ManagerFactory.  The first time it is
	 * called, the single instance is created.
	 *
	 * @return the singleton instance
	 */

	static ManagerFactory* getInstance();

	/**
	 * This method creates a manager given the appropriate URI.
	 * Any additional configuration of the manager will be passed
	 * in via the URI in the form of manager-specific key/value
	 * pairs at the end of the URI.  Each URI will have as the
	 * protocol the specific manager that should be created.
	 *
	 * @param uri the desired manager
	 * @return the manager instance
	 * @throws FatalException if the manager can't be created
	 */

	virtual AccountManager* createManager(const std::string& uri)
				throw(FatalException);

	/**
	 * This method returns an iterator pointer to access all of
	 * the registered protocols.
	 *
	 * @return the iterator pointer
	 */

	virtual iterator* elements();

	/**
	 * This method is used to register new manager creator classes
	 * with the factory.
	 */

	virtual void registerCreator(const std::string& s, ManagerCreator *c);

protected:
	ManagerFactory() {}

	friend class auto_ptr<ManagerFactory>;
	virtual ~ManagerFactory() {}

private:
	typedef std::list<std::string>			string_list;
	typedef std::map<std::string, ManagerCreator*>	creator_map;

	static string_list			_list;
	static creator_map			_map;
	static ManagerFactory*			_instance;
	static auto_ptr<ManagerFactory>		_del;
};

#endif
