///////////////////////////////////////////////////////////////////////
//
// This file is part of the finlink conduit for PalmCoins 
// Copyright (C) 2001 Andrew S. Townley 
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA,
// 02111-1307  USA
//
// Name:	ErrorTrace.cpp
// Author:	Andrew S. Townley <adz1092 at netscape dot net>
// Created:	Mon Jun 11 18:46:22 CDT 2001
//
///////////////////////////////////////////////////////////////////////

#include "ErrorTrace.h"

string ErrorTrace::_traceFileName;
FILE* ErrorTrace::_traceFile = static_cast<FILE*>(NULL);
int ErrorTrace::_traceLevel = 0;

const char* ErrorTrace::TRACE_START_FMT = "%s()";
const char* ErrorTrace::TRACE_RETURN_FMT = "%s() returning";
const char* ErrorTrace::TRACE_RETURN_INT_FMT = "%s() returning:  %d";
const char* ErrorTrace::TRACE_RETURN_PTR_FMT = "%s() returning:  %p";
const char* ErrorTrace::TRACE_RETURN_TRC_FMT = "%s() returning:  %s";
const char* ErrorTrace::TRACE_RETURN_STR_FMT = "%s() returning:  '%s'";
const char* ErrorTrace::TRACE_RETURN_DBL_FMT = "%s() returning:  %lf";
const char* ErrorTrace::TRACE_ARG_STR_FMT = "\t%s = '%s'";
const char* ErrorTrace::TRACE_ARG_INT_FMT = "\t%s = %d";
const char* ErrorTrace::TRACE_ARG_PTR_FMT = "\t%s = %p";
const char* ErrorTrace::TRACE_ARG_TRC_FMT = "\t%s = %s";
const char* ErrorTrace::TRACE_ARG_DBL_FMT = "\t%s = %lf";
const char* ErrorTrace::TRACE_THROW_FMT = "%s() throwing exception %s";
const char* ErrorTrace::TRACE_THROW_MSG_FMT = "%s() throwing exception %s with message:  %s";

ErrorTrace::ErrorTrace()
	: _maturity(0), _className("[unset]")
{
}

ErrorTrace::ErrorTrace(const string& className, int maturity)
{
	init(className, this, maturity);
}

void ErrorTrace::init(const string& className, void *ptr, int maturity)
{
	char buf[256];
	snprintf(buf, 256, "%s[%p]", className.c_str(), ptr);
	_className = buf;
	_maturity = maturity;
}

ErrorTrace::~ErrorTrace()
{
}

void ErrorTrace::methodStart(const char *name)
{
	tprintf(1, TRACE_START_FMT, name);
}

void ErrorTrace::methodArg(const char *name, const char *arg)
{
	tprintf(1, TRACE_ARG_STR_FMT, name, arg);
}

void ErrorTrace::methodArg(const char *name, const string& arg)
{
	tprintf(1, TRACE_ARG_STR_FMT, name, arg.c_str());
}

void ErrorTrace::methodArg(const char *name, int arg)
{
	tprintf(1, TRACE_ARG_INT_FMT, name, arg);
}

void ErrorTrace::methodArg(const char *name, void *arg)
{
	tprintf(1, TRACE_ARG_PTR_FMT, name, arg);
}

void ErrorTrace::methodArg(const char *name, bool arg)
{
	tprintf(1, TRACE_ARG_TRC_FMT, name, (arg ? "true" : "false"));
}

void ErrorTrace::methodArg(const char *name, double arg)
{
	tprintf(1, TRACE_ARG_DBL_FMT, name, arg);
}

void ErrorTrace::methodArg(const char *name, Traceable& arg)
{
	tprintf(1, TRACE_ARG_TRC_FMT, name, arg.toString().c_str());
}

void ErrorTrace::methodArg(const char *name, Traceable *arg)
{
	if(arg != static_cast<Traceable*>(NULL))
		{
		tprintf(1, TRACE_ARG_TRC_FMT, name, arg->toString().c_str());
		}
	else
		{
		tprintf(1, TRACE_ARG_PTR_FMT, name, arg);
		}
}

void ErrorTrace::methodReturn(const char *name)
{
	tprintf(1, TRACE_RETURN_FMT, name);
}

//void ErrorTrace::methodReturn(const char *name, bool arg)
//{
//	tprintf(1, TRACE_RETURN_TRC_FMT, name, (arg ? "true" : "false"));
//}

void ErrorTrace::methodReturn(const char *name, void *arg)
{
	tprintf(1, TRACE_RETURN_PTR_FMT, name, arg);
}

void ErrorTrace::methodReturn(const char *name, const char *arg)
{
	tprintf(1, TRACE_RETURN_STR_FMT, name, arg);
}

void ErrorTrace::methodReturn(const char *name, const string& arg)
{
	tprintf(1, TRACE_RETURN_STR_FMT, name, arg.c_str());
}

void ErrorTrace::methodReturn(const char *name, double arg)
{
	tprintf(1, TRACE_RETURN_DBL_FMT, name, arg);
}

void ErrorTrace::methodReturn(const char *name, Traceable& arg)
{
	tprintf(1, TRACE_RETURN_TRC_FMT, name, arg.toString().c_str());
}

void ErrorTrace::methodReturn(const char *name, Traceable *arg)
{
	if(arg != static_cast<Traceable*>(NULL))
		{
		tprintf(1, TRACE_RETURN_TRC_FMT, name, arg->toString().c_str());
		}
	else
		{
		tprintf(1, TRACE_RETURN_PTR_FMT, name, arg);
		}
}

void ErrorTrace::methodThrow(const char *name, const char *except, const char *msg = static_cast<const char*>(NULL))
{
	if(msg != static_cast<const char*>(NULL))
		tprintf(1, TRACE_THROW_FMT, name, except, msg);
	else
		tprintf(1, TRACE_THROW_MSG_FMT, name, except);
}

void ErrorTrace::tprintf(int threshold, const char *fmt, ...)
{
	if(!willTrace(threshold))
		return;

	fprintf(_traceFile, _className.c_str());
	fprintf(_traceFile, " ");
	va_list	args;
	va_start(args, fmt);
	vfprintf(_traceFile, fmt, args);
	va_end(args);
	fprintf(_traceFile, "\n");
	fflush(_traceFile);
}

bool ErrorTrace::willTrace(int threshold)
{
	if(threshold >= ((_maturity * 10) + _traceLevel)
			&& _traceFile != NULL)
		{
		return true;
		}
	
	return false;
}

void ErrorTrace::setTraceLevel(int l)
{
	_traceLevel = l;
}

void ErrorTrace::setTraceFile(const string& filename)
{
	if(_traceFile != NULL)
		{
		fclose(_traceFile);
		}
	_traceFile = fopen(filename.c_str(), "w+");
	if(_traceFile != NULL)
		{
		_traceFileName = filename;
		}
}
