///////////////////////////////////////////////////////////////////////
//
// This file is part of the finlink conduit for PalmCoins 
// Copyright (C) 2001 Andrew S. Townley 
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA,
// 02111-1307  USA
//
// Name:	finlink.cpp
// Author:	Andrew S. Townley <adz1092 at netscape dot net>
// Created:	Sat Jun  2 18:20:03 CDT 2001
//
///////////////////////////////////////////////////////////////////////

#include <popt.h>
#include "finlink.h"

using namespace std;

static int	op_copy_from = 0;
static int	op_copy_to = 0;
static int	op_merge = 0;
static int	op_dump = 0;
static int	op_list = 0;
static string	from_uri;
static string	to_uri;

static struct poptOption optionsTable[] =
{
	{ "source", 's', POPT_ARG_STRING, NULL, 's', "specify the sync source", "source-uri" },
	{ "dest", 'd', POPT_ARG_STRING, NULL, 'd', "specify the sync destination", "target-uri" },
	{ "copy-from", 'f', POPT_ARG_NONE, &op_copy_from, 'f', "copy data from the source to the destination", "" },
	{ "copy-to", 't', POPT_ARG_NONE, &op_copy_to, 't', "copy data from the destination to the source", "" },
	{ "merge", 'm', POPT_ARG_NONE, &op_merge, 'm', "merge data from the source and the destination (synchronize)", "" },
	{ "list", 'l', POPT_ARG_NONE, &op_list, 'l', "list all registered modules", "" },
	{ "dump", 0, POPT_ARG_NONE | POPT_ARGFLAG_DOC_HIDDEN, &op_dump, 'p', "dump the data to stdout, but don't modify source or target", "" },
	POPT_AUTOHELP
	{ 0, 0, 0, 0, 0, 0, 0 }
};

// prototypes
void check_options(poptContext optCon);
void copy();
void do_copy(AccountManager *src, AccountManager *dest);
void do_merge(AccountManager *src, AccountManager *dest);
//void dump_accounts(ostream& s, AccountManager *mgr, bool name_only);
void dump_accounts(FILE* s, AccountManager *mgr, bool name_only);
//void dump_helper(ostream& s, AccountNode::iterator* i, bool name_only);
void dump_helper(FILE* s, AccountNode::iterator* i, bool name_only);
//void dump_transactions(ostream& s, AccountNode *n);
void dump_transactions(FILE* s, AccountNode *n);
void list_modules();
void load_transactions(AccountManager *m, AccountNode *dn, AccountNode *n);
void merge();
void usage(poptContext optCon, int exitcode, char *error, char *addl);
ostream& operator<<(ostream& os, AccountNode *n);

int main(int argc, const char *argv[])
{
	char		c;
	int		i = 0;
	string		op;
	poptContext	optCon;

	ErrorTrace::setTraceLevel(1);
	ErrorTrace::setTraceFile("finlink.trace");

	optCon = poptGetContext(NULL, argc, argv, optionsTable, 0);
	if(argc == 1)
		{
		usage(optCon, -1, "error", 
			"must specify source, target and operation");
		}

	// process the options
	while((c = poptGetNextOpt(optCon)) > 0)
		{
		switch(c)
			{
			case 's':
				from_uri = poptGetOptArg(optCon);
				if(from_uri.length() == 0)
					{
					usage(optCon, -2, "error",
						"source URI cannot be zero length");
					}
				break;
			case 'd':
				to_uri = poptGetOptArg(optCon);
				if(to_uri.length() == 0)
					{
					usage(optCon, -3, "error",
						"target URI cannot be zero length");
					}
				break;
			case 'f':
				if(op_copy_to || op_merge)
					{
					usage(optCon, -4, "error",
						"only one of copy-from, copy-to or merge may be specified");
					}
				op = "copy from source";
				break;
			case 't':
				if(op_copy_from || op_merge)
					{
					usage(optCon, -5, "error",
						"only one of copy-from, copy-to or merge may be specified");
					}
				op = "copy to source";
				break;
			case 'm':
				if(op_copy_from || op_copy_to)
					{
					usage(optCon, -6, "error",
						"only one of copy-from, copy-to or merge may be specified");
					}
				op = "synchronize";
				break;
			}
		}

	if(c < -1)
		{
		cerr << "error: " << poptStrerror(c) << " '";
		cerr << poptBadOption(optCon, POPT_BADOPTION_NOALIAS);
		cerr << "'" << endl;
		usage(optCon, -7, NULL, NULL);
		}

	// final complicated check of options
	check_options(optCon);
	
	// done with options
	poptFreeContext(optCon);

	// load the registered modules
	ModuleLoader *ml = ModuleLoader::getInstance();
	ml->loadModules();

	if(op_list)
		{
		list_modules();
		}
	else if(op_copy_from || op_copy_to)
		{
		copy();
		}
	else if(op_merge)
		{
		merge();
		}

	exit(0);
}

/**
 * This function is responsible for making sure that the usage is
 * correct beyond the simple sanity checking that is doine in the
 * options parsing code above.
 */

void check_options(poptContext optCon)
{
	// 1.  make sure that the source and dest aren't empty
	if((to_uri.length() != 0 && from_uri.length() == 0) ||
			(from_uri.length() != 0 && to_uri.length() == 0))
		{
		usage(optCon, -8, "error",
			"must specify both source and destination URI");
		}
	// 2.  make sure we've got URIs if we have an operation
	if((op_copy_from || op_copy_to || op_merge) &&
		(to_uri.length() == 0 && from_uri.length() == 0))
		{
		usage(optCon, -9, "error",
			"operations must include source and destination URI");
		}
}

/**
 * This function takes care of the entire copy process.
 */

void copy()
{
	AccountManager *src;
	AccountManager *dest;

	ManagerFactory *mf = ManagerFactory::getInstance();
	if(op_copy_from)
		{
		try
			{
			src = mf->createManager(from_uri);
			}
		catch(const FatalException& e)
			{
			cerr << e.getMessage() << endl;
			exit(1);
			}
		try
			{
			dest = mf->createManager(to_uri);
			}
		catch(const FatalException& e)
			{
			delete src;
			cerr << e.getMessage() << endl;
			exit(2);
			}
		}
	else if(op_copy_to)
		{
		try
			{
			src = mf->createManager(from_uri);
			}
		catch(const FatalException& e)
			{
			cerr << e.getMessage() << endl;
			exit(1);
			}
		try
			{
			dest = mf->createManager(to_uri);
			}
		catch(const FatalException& e)
			{
			delete src;
			cerr << e.getMessage() << endl;
			exit(2);
			}
		}

	do_copy(src, dest);

	delete src;
	delete dest;
}

/**
 * This function takes care of handling the copy operation between the
 * two account managers.
 *
 * @param src the source account manager
 * @param dest the dest account manager
 */

void do_copy(AccountManager *src, AccountManager *dest)
{
	try
		{
		dest->deleteAllAccounts();
		src->readAccounts();

		AccountNode *nn;

		// pass 1:  copy all accounts
		AccountNode::iterator_ptr p(src->accounts());
		for(p->first(); p->moreElements(); p->next())
			{
			nn = dest->newNode();
			dest->addAccount(nn);
			nn->copy(p->current());
			}

		// pass 2:  copy the transactions
		for(p->first(); p->moreElements(); p->next())
			{
			load_transactions(dest, nn, p->current());
			}

		dump_accounts(stdout, src, true);
		dump_accounts(stderr, dest, true);
		dest->writeAccounts();
		}
	catch(const FatalException& e)
		{
		cerr << e.getMessage() << endl;
		}
}

/**
 * This function takes care of handling the merge operation between
 * the two account managers.
 *
 * @param src the source account manager
 * @param dest the destination account manager
 */

void do_merge(AccountManager *src, AccountManager *dest)
{
}

/**
 * This function is used to dump the accounts for a given account
 * manager.
 *
 * @param s the stream to which the accounts will be written
 * @param mgr the manager to dump
 * @param name_only true if only the account name should be printed
 */

//void dump_accounts(ostream& s, AccountManager *mgr, bool name_only)
void dump_accounts(FILE* s, AccountManager *mgr, bool name_only)
{
	dump_helper(s, mgr->accounts(), name_only);
}

/**
 * Recursive function to assist in walking the tree.
 */

//void dump_helper(ostream& s, AccountNode::iterator* i, bool name_only)
void dump_helper(FILE* s, AccountNode::iterator* i, bool name_only)
{
	IteratorPtr<AccountNode *> p(i);
	for(p->first(); p->moreElements(); p->next())
		{
		AccountNode *n2 = p->current();
		if(name_only)
			{
			fprintf(s, "%s\n", n2->getPath("::").c_str());
			dump_transactions(s, n2);
			fprintf(s, "\n");
			}
		else
			{
//			s << n2;
			}

		// process children
		dump_helper(s, n2->children(), name_only);
		}
}

/**
 * This method will dump the transactions for a given account.
 */

//void dump_transactions(ostream& s, AccountNode *n)
void dump_transactions(FILE* s, AccountNode *n)
{
	// FIXME: I know, I know... but I can't remember how to do
	// this and my books are all packed!!!
	FILE *fd = s;
	
	TransactionNode::iterator_ptr p(n->transactions());
	for(p->first(); p->moreElements(); p->next())
		{
		TransactionNode	*tn = p->current();
		fprintf(fd, "%10s %-6s %-38s  |%10.2lf|%10.2lf\n",
			Format::fmt_date(tn->getDate()).c_str(),
			tn->getNumber().c_str(),
			Format::elid_string(tn->getDescription(), 38).c_str(),
			tn->getValue(),
			tn->getBalance());
		}
	fprintf(fd, "Ending balance:  %10.2lf\n", n->getBalance());
}

/**
 * This function lists all the registered modules and the
 * module-specific URI information.
 */

void list_modules()
{
	cout << "Registered modules:" << endl << endl;

	ManagerFactory *mf = ManagerFactory::getInstance();
	ManagerFactory::iterator_ptr p(mf->elements());
	for(p->first(); p->moreElements(); p->next())
		{
		string base = p->current();
		// FIXME:  need to remember how to do this with
		// iostreams!!
		printf("\t%-15s", base.c_str());
		try
			{
			AccountManager *mgr = mf->createManager(base);
			cout << "Usage: " << mgr->getUsage();
			delete mgr;
			}
		catch(const FatalException& e)
			{
			; // ignore it
			}
		cout << endl;
		}
	cout << endl;
}

void load_transactions(AccountManager *m, AccountNode *dn, AccountNode *n)
{
	// load the transactions for this node
	TransactionNode::iterator_ptr tp(n->transactions());
	for(tp->first(); tp->moreElements(); tp->next())
		{
		TransactionNode *cn = tp->current();

		// look to see if we've already entered this
		// transaction

cerr << "Transaction:  " << cn->getDescription() << endl;

		TransactionNode *nn = m->getTransaction(cn);
		if(nn == static_cast<TransactionNode*>(NULL))
			{
cerr << "\tdidn't have it, so created a new one" << endl;
			nn = dn->newTransaction();
			nn->copy(cn, m);
			m->addTransaction(nn);
{
cerr << "SOURCE TRANS:  '" << cn->getNumber() << "'" << endl;
cerr << "=====================================================" << endl;
TransSplit::iterator_ptr p(cn->splits());
for(p->first(); p->moreElements(); p->next())
	{
	TransSplit *ts = p->current();
	cerr << "Split - " << endl;
	cerr << "*** account:  " << ts->getAccount() << endl;
	cerr << "*** memo:     " << ts->getMemo() << endl;
	cerr << "*** value:    " << ts->getValue() << endl;
	}
cerr << "=====================================================" << endl;
cerr << "TARGET TRANS:  '" << nn->getNumber() << "'" << endl;
cerr << "=====================================================" << endl;
TransSplit::iterator_ptr q(nn->splits());
for(q->first(); q->moreElements(); q->next())
	{
	TransSplit *ts = q->current();
	cerr << "Split - " << endl;
	cerr << "*** account:  " << ts->getAccount() << endl;
	cerr << "*** memo:     " << ts->getMemo() << endl;
	cerr << "*** value:    " << ts->getValue() << endl;
	}
cerr << "=====================================================" << endl;
}
			}
		else
			{
cerr << "\tfound it; doing nothing" << endl;
			}
		}

	// process the children
	AccountNode::iterator_ptr p(n->children());
	for(p->first(); p->moreElements(); p->next())
		{
		load_transactions(m, dn, p->current());
		}
}

/**
 * This function handles the merge operation between the source and
 * the destination.
 */

void merge()
{
}

void usage(poptContext optCon, int exitcode, 
		char *error, char *addl)
{
	if(error)
		{
		cerr << error << ": " << addl << endl;
		}
	poptPrintUsage(optCon, stdout, 0);
	poptFreeContext(optCon);
	exit(exitcode);
}

ostream& operator<<(ostream& os, AccountNode *n)
{
	string	s;

	switch(n->getType())
		{
		case AccountNode::INCOME:
			s = "Income";
			break;
		case AccountNode::EXPENSE:
			s = "Expense";
			break;
		default:
			s = "Unknown";
			break;
		}
	
	os << "Account:         " << n->getPath("::") << endl;
	os << "Type:            " << s << endl;
	os << "Balance:         " << n->getBalance() << endl;
	os << "Description:     " << n->getDescription() << endl;
	os << endl;

	return os;
}

