/**	
	Code to access the database containing the account information

	$Revision: 1.35 $
	Last updated: $Date: 2001/07/24 01:00:04 $
	
	Copyright (C) 2000-2001Vlad Mereuta <dizzy@users.sourceforge.net>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
	
	$Log: AccountsDB.h,v $
	Revision 1.35  2001/07/24 01:00:04  chaese
	eliminated references to PalmCoins except in situations that require
	renaming CVS controlled files (or other instances that I've overlooked :)
	
	Revision 1.34  2001/07/23 10:48:33  chaese
	Pack bools into bitfields
	
	Revision 1.33  2001/06/29 16:17:26  dizzy
	Expanded Carsten's #define to handle strings and transformed it into a template fct. This should help tidying up packing operations a bit
	(provided it works as it should)
	
	Revision 1.32  2001/06/29 00:24:57  chaese
	Rewrote how total balances are tracked. This should eliminate all problems
	that were associated with the old way (i.e. balances not being recalculated
	when they should have been etc.)
	
	Revision 1.31  2001/06/27 22:58:11  dizzy
	Restructured #include directives
	
	Revision 1.30  2001/06/27 03:15:12  chaese
	Resurrected category selector in main form as type group selector
	
	Revision 1.29  2001/06/21 11:04:23  chaese
	* initial support for account types
	* fixed some obscure bugs that materialized on a PalmOS 3.5 debug rom
	(see related message on palmcoins-misc for details)
	
	Revision 1.28  2001/06/17 05:23:32  chaese
	*** CAUTION: DATABASE STRUCTURE HAS CHANGED YET AGAIN ***
	I generally don't like 'cross-logging', but too many changes have accumulated
	to keep them separated.
	1. completely eliminated account categories
	2. introduced currently unused type field in the account record structure
	(This will become what the user sees as 'category selector'. Coming soon!)
	3. completely eliminated protected accounts (mostly because #1 made it so much
	harder to locate the "Misc" account in the Expense category:) Those should be
	replaced with Pref. settings for default Income and Expense accounts anyway.
	4. changes to user interface; some as mere consequences of 1. (e.g.
	Account chooser), some for usability reasons (Transaction form).
	5. bugfixes and minor improvements.
	
	Revision 1.27  2001/06/14 00:52:05  chaese
	show account names with as much of their full path name as possible
	
	Revision 1.26  2001/06/10 04:05:41  chaese
	basic infrastructure for split transactions (somewhat tested, but not
	guaranteed to be bug free)
	CAUTION! Database format has changed! Expect nastiness when you run the
	new version on an old Database.
	
	Revision 1.25  2001/06/06 00:59:11  chaese
	Bugfix: Reparenting accounts now keeps track of balances correctly.
	
	Revision 1.24  2001/06/06 00:09:52  chaese
	Bugfix and code cleanup regarding the adjustment of parent balances. This
	fixes the bugs that deleting a transaction and adjusting a balance didn't
	properly update the parent balances, and removes a bit of code duplication.
*/

#ifndef __ACCOUNTS_DB__
#define __ACCOUNTS_DB__


#include "Database.h"
#include "MiscFunc.h" //for section definitions

/// type of the account database
#define ACCOUNTSDB_TYPE	'DATA'
/// db name
#define ACCOUNTSDB_NAME	"FreeCoins-Accounts"

///this structure contains the overall totals for all the selected accounts
typedef struct {
	///current balance total
	Int32	total;
	///end of month total
	Int32	expected_total;
} totalsStruct;

/// sort order type
typedef enum { soName, soAmount, soDate } sortOrder;

/// account flags for tight packing
typedef struct {
	unsigned collapsed		:1;
	unsigned credit_account	:1;
	unsigned reserved		:14;
} AccountFlags;

///the packed structure of an account - see the unpacked struct for explanations
typedef struct {
	Int32		balance;
	Int32		expected_balance;
	Int32		credit_limit;
	UInt32		parent_uid;
	Int32		children_total;
	Int32		children_expected_total;
	AccountFlags flags;
	UInt8		acc_type;
	UInt8		depth;
	UInt8		children_no;
	Char		description[1]; // actually may be longer than 1
} packedAccountStructure;

enum AccountType { BANK=0, CREDIT=1, INCOME=2, EXPENSE=3 };

///unpacked structure of an account
typedef struct {
	/// account balance
	Int32		balance;
	/**
		expected balance in the account at a set date (after transactions cleared and 
		standing orders/direct debits which were due went out)
	*/
	Int32		expected_balance;
	///the credit limit for creditcards/overdraft for other accounts
	Int32		credit_limit;	
	/// uid of the parent of this account. if the account is a root one this field should be ignored
	UInt32		parent_uid;
	/// overall balance of the children of this account
	Int32		children_total;
	/// overall expected balance of the children of this account
	Int32		children_expected_total;
	/// account type (Bank, CC, etc.)
	AccountType	acc_type;
	/// depth at which this account is situated (0 for root accounts, >0 for the others)
	UInt8		depth;
	/// total number of children owned by this account. this should facilitate things like sorting
	UInt8		children_no;	
	/// indicates whether the account is collapsed or not
	UInt8		collapsed;
	/// if true, the program will credit money to this account when spending them and subtract from the balance when paying money into it
	UInt8		credit_account;
	///account desciption
	Char* 		description;
	///the name of the account - this will show up throughout the program
	Char*		acname;
} accountStructure;

/// class to manipulate the accounts database
class AccountsDB : public Database <accountStructure, packedAccountStructure>
{
private:
	UInt32 createAccount(char *name, UInt32 parent_uid, AccountType acc_type);

 public:
	totalsStruct	totals;
		
 	/**
		default constructor
		opens or creates the database (if created populates it with sample records)
	*/
 	AccountsDB() ACDB_SECTION;
	
	/**
		constructor
		@param db_open_ref reference to an *already* opened database
	*/
	AccountsDB(DmOpenRef db_open_ref) ACDB_SECTION;
	 
	/**
		@param n name of the account to be found
		@param categ category of the account
		@return unique id of the account or 0 if the account could not be found
	*/
	UInt32	findAccountByName(char* n, UInt16 categ) ACDB_SECTION;
	
	/**
		returns a formated account name in buf
		@param acc account to get the name for
		@param buf an *allocated* buffer in which the name is to be copied
	*/
	void	getAccountName(UInt32 acc, Char* buf) ACDB_SECTION;

	/**
		Build the account's full path name, truncating the beginning if
		necessary to fit the given number of pixels.
		@param acc The account's UID
		@param width number of pixels
		@param dest allocated buffer to receive the result. I figure that it
			should be able to store width/2 characters in the worst case.
	*/
	void	getFullNameTruncLeftToFit(UInt32 acc, Int16 width, Char *dest)
		ACDB_SECTION;

	/// unpacks an account record
	void	unpackRecord (accountStructure* account, packedAccountStructure* packedAccount) ACDB_SECTION;

	/// packs an account record
	void	packRecord (accountStructure* account, MemHandle accountDBEntry) ACDB_SECTION;
	
	/**
		Increment the account's balances by specified amounts, notifying the
		ancestors about the change as well.
		This is a 'lowlevel' function that modifies the account balance
		directly.
		@param uid uid of the account to be affected
		@param clr_amt amount to be added to the cleared balance
		@param exp_amt amount to be added to the expected balances
		@param is_parent true if the uid is the parent of an account whose
			balance has already changed. In this case, only the children totals
			of the given account and its ancestors are updated.
	*/
	void incrementBalance(UInt32 uid, Int32 clr_amt, Int32 exp_amt,
			Boolean is_parent = false)  ACDB_SECTION;

	/**
		Updates the balance for an account. This is a 'highlevel' function
		that affects the balance by inserting a cleared balance adjustment
		transaction.
		@param uid unique id
		@param newbal the new balance
		@param payee optional payee description
	*/
	void updateBalance(UInt32 uid,Int32 newbal,char *payee=NULL) ACDB_SECTION;

	/**
		fetch the balance of a given account
		@param id	the unique id of the account record
		@return the balance in pence
	*/
	Int32	getBalance(UInt32 uid) ACDB_SECTION;

	/**
		fetch the expected balance of a given account
		@param id	the unique id of the account record
		@return the expected balance in pence
	*/
	Int32	getExpectedBalance(UInt32 uid) ACDB_SECTION;

	/**
		checks if a parent of the current account is collapsed
		@acc account id
		@return true if no parrent is collapsed, false otherwise
	*/
	Boolean	isDisplayable(UInt16 acc) ACDB_SECTION;
	
	//init the Accounts DB with some default records
	void	initializeAccounts(void) ACDB_SECTION;

	/**
		packs and writes the data from s into the database
		note: after calling this function the record has to be released

		@return the index to the record
	*/
	UInt16	newRecord(accountStructure* s) ACDB_SECTION;

	/**
		Deletes an account given by its id
		The function scans through all the transactions and removes the ones which
		have as source/destinations the account to be deleted and as dest/src an account
		in the Income or Expense categories.

		All the transaction involving user defined account will be updated so that the 
		reference to the account to be deleted is replaced with a reference to a Income/Expense
		account

		@param id the id of the account to be deleted
	*/
	void	deleteRecord(UInt16	id) ACDB_SECTION;

	/**
		Compares two records (by the account name)
	*/
	static Int16	compareRecords(packedAccountStructure* rec1,packedAccountStructure* rec2,Int16 other,SortRecordInfoPtr sr1,SortRecordInfoPtr
			sr2,MemHandle app_info) ACDB_SECTION;

	/// recategorize accounts as visible/invisible depending on user's choice
	void updateVisibility(void);
};

/**
	unfortunatelly this is a bit messy :(. We need to be able to access this db from anywhere in the program (including the compareRecords function).
	The best solution I managed to come up with is defining the pointer in .c file, and exporting it in .h (tried defining it as static in .h only but
	did not work. Next we need to instatntiate the class (this is done in palmcoins.c) and set the pointer to a valid address
	
*/
extern AccountsDB* accounts_db;
#endif
