/**	
	$Revision: 1.10 $
	Last updated: $Date: 2001/06/29 21:32:44 $

	Copyright (C) 2000-2001 Vlad Mereuta <dizzy@users.sourceforge.net>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
		
	$Log: Database.h,v $
	Revision 1.10  2001/06/29 21:32:44  dizzy
	Eliminated bug discovered with os4 (releasing record which was not busy)
	
	Revision 1.9  2001/06/29 16:17:26  dizzy
	Expanded Carsten's #define to handle strings and transformed it into a template fct. This should help tidying up packing operations a bit
	(provided it works as it should)
	
	Revision 1.8  2001/06/27 22:58:11  dizzy
	Restructured #include directives
	
*/

#ifndef __DATABASE_H__
#define __DATABASE_H__

#include <PalmOS.h>
#include <FeatureMgr.h>

/// Application Info Block
typedef struct {
	UInt16	renamedCategories;					// bitfield of categories with a different name
	Char	categoryLabels [dmRecNumCategories] [dmCategoryLength];
	UInt8	categoryUniqIDs [dmRecNumCategories];
	UInt8	lastUniqID;							// Uniq IDs generated by the device are between
												// 0 - 127.  Those from the PC are 128 - 255.
	UInt8	reserved1;							// from the compiler word aligning things
	UInt16	reserved2;	
	UInt16	dirtyAppInfo;
	UInt8	sortOrder;
	UInt8	reserved3;
} appInfoType;


/**
	interface to some of the PalmOS database functions
	Using this class:
	* derive a new database class as public from this one passing your packed/unpacked structures as parameters.
	* in your new class override the pure virtual function from this class and add anything you might thing useful		
*/
template <class recordStructure, class packedRecordStructure> class Database
{
 protected: 
	/// reference to the database
	DmOpenRef	ref;
	/// total number of categories
	UInt32		num_cat;
	/// true if the db had to be created when it was opened
	Boolean		created;
	///true if a record is selected, false otherwise
	Boolean		record_selected;
	/// the current record
	UInt16		selected_record;

	/**
		opens or creates a database
		@param type database type
		@param creator database creator
		@param mode	mode in which to open the database
		@param name db name
		@param created on return is set to true if the db had to be created
	*/
	Err		openOrCreate(UInt32 type, UInt32 creator, UInt16 mode, UInt16 cardNo, Char* name);

	///returns a pointer to the app info record
	MemPtr	getLockedAppInfo();

	/**
		returns a pointer to the app info record
		@param categ_string the function will use this id to initialise the categories for the db
	*/
	MemPtr	getLockedAppInfo(UInt16 categ_string);

	///used by the func above to initialise the app info rec if it is empty
	MemPtr appInfoInit(LocalID dbID, UInt16 cardNo);

	///used by the func above to initialise the app info rec if it is empty
	MemPtr appInfoInit(LocalID dbID, UInt16 cardNo, UInt16 categ_string);
		 
 public:
	/// the currenct category
	UInt16		current_cat;

	/**
		get the pointer to the db (to use for calling PalmOS db functions directly)
		@return  pointer to the db (DmOpenRef)
	*/
	DmOpenRef getRef() { return ref; }
	
	/// @return true if the constructor created the db instead of just opening it
	Boolean isNew() { return created; }
	
	/// @return the uid of the selected_record
	UInt32	getSelectedUid(); 
 
	/**
		Fetch the uid of a record
		@param id	the record id

		@return the record uid
	*/
	UInt32	getRecordUid(UInt16 id);

	/**
		Fetch the uid of the current record
		@return the record uid
	*/
	UInt32	getRecordUid();

	/**
		Fetch the uid of a record
		@param uid	the record uid

		@return the record id
	*/
	UInt16	getRecordId(UInt32 uid);

	/**
		Returns the category of a given record
		@param id	the id of the record
		@return	the record category
	*/
	UInt16	getRecordCategory(UInt16 id);

	/**
		Returns the category of a given record
		@param index	unique id of the record
		@return	the record category
	*/
	UInt16	getRecordCategory(UInt32 uid);

	/** 
		sets the category of a given record
		@param id	id of the record
		@param cat		the record category
	*/
	void	setRecordCategory(UInt16 id, UInt16 cat);

 	/** 
		sets the category of a given record
		@param uid	unique id of the record
		@param cat	the record category
	*/
	void	setRecordCategory(UInt32 uid, UInt16 cat);

	/**
		finds a category in this db by its name
		@param cat_name the name of the category to be found
		@return the category id
	*/
	UInt16	findCategory(Char* cat_name);
	
	/**
		fetches the name of a category in an allocated buffer
		@param cat_id the id of the category (in the current db only)
		@param cat_name allocated buffer to which the name will be written
	*/
	void	getCategoryName(UInt16 cat_id, Char* cat_name);
	
	/**
		get the unpacked strucure of a given record
		@param id	id of the account record
		@param rec 	pointer in which the unpacked data will be saved
		@return pointer to the record handler (of the packed record)
	*/
	MemHandle getUnpackedRecord(UInt16 id,recordStructure* rec);

	/**
		get the unpacked strucure of the current record
		@param rec 	pointer in which the unpacked data will be saved
		@return pointer to the record handler (of the packed record)
	*/
	MemHandle getUnpackedRecord(recordStructure* rec);
	
	/**
		get the unpacked strucure of a given record
		@param uid	unique id of the account record
		@param rec 	pointer in which the unpacked data will be saved
		@return pointer to the record handler (of the packed record)
	*/
	MemHandle getUnpackedRecord(UInt32 uid,recordStructure* rec);

	/**
		sets the selected_record to id and sets record_selected flag to false
		@param id id of the record to be selected
	*/
	void	selectRecord(UInt16 id);

	/**
		sets the selected_record from an unique and sets record_selected flag to false
		@param id id of the record to be selected
	*/
	void	selectRecord(UInt32 uid);

	/**
		if a record is selected this function unselects it (by setting the record_selected flag)
	*/
	void	unselectRecord() { record_selected=false; }	
	
	/**
		use this function to find out whether a record from the database is currently selected
		@return true if a record is selected, false otherwise
	*/
	Boolean	isRecordSelected() { return record_selected; }

	/**
		use this function to retrieve the currently selected record
		@return the index of the selected record
	*/
	UInt16	getSelectedRecord();
	
	/**
		constructor
		use when in need to access functions like pack/unpack from a static function.
		@param db_open_ref reference to an *already* opened database
	*/
	Database(DmOpenRef db_open_ref);

	/**
		initialize the databases
		@param creator database creator
		@param db_name name of the database 
		@param type type of the database (ex. 'DATA')
	*/
	Database(UInt32 type, UInt32 creator, Char* db_name);

	/**
		initialize the databases
		@param type type of the database (ex. 'DATA')
		@param creator database creator
		@param db_name - name of the database
		@param categ_string - id of the category resource to be used to initialise the categories for this database
	*/
	Database(UInt32 type, UInt32 creator, Char* db_name, UInt16 categ_string);

	/// closes the database
	virtual ~Database();

	/**
		unpacks a record
		@param rec the record pacxked_rec will be unpacked into this location
		@param packed_rec pointer to the record to be unpacked
	*/
	virtual void	unpackRecord (recordStructure* rec, packedRecordStructure* packed_rec)=0;

	/**
		packs an account record. This function should be overloaded by all database classes
		@param rec pointer to the record to be packed
		@param db_entry pointer to the newly allocated packed record
	*/
	virtual void	packRecord (recordStructure* rec, MemHandle db_entry)=0;
	
	/**
		This function writes a given string into a structure; it is meant as a helper function for Database#packRecord. Also see 
		Database#writeAndAdvance.
		
		@note	I am actually amazed that C++ can handle polymorphism in this way...		
		
		@param s pointer to the start of the block where the str is to be written
		@param offset offset within the block. note that offset is changed by the function
		@param str string to be written
	*/	
	void writeAndAdvance(MemPtr s, UInt32& offset, Char* str)
	{
			DmStrCopy(s, offset, str);
			offset += StrLen(str)+1;		
	}
	
	/**
		This is a helper function for pack record. Given a field from a structure together with a destination pointer and an offset
		it will write the field in the specified location. 
		@param s pointer to the start of the block where the str is to be written
		@param offset offset within the block. note that offset is changed by the function
		@param f a \b pointer to the field to be written. if it is a char* it will (hopefully) be handled by the other writeAndAdvance funcition
		
	*/
	template <class Field> void writeAndAdvance(MemPtr s, UInt32& offset, Field* f)
	{
			DmWrite(s, offset, (MemPtr) f, sizeof(Field));
			offset += sizeof(Field);
	}
		
	/**
		packs and writes the data from rec at the end of the database.
		@return the index to the record
	*/
	virtual UInt16	newRecord(recordStructure* rec);

	/**
		Deletes a record given by its id
		@param id the id of the record to be deleted
	*/
	virtual void deleteRecord(UInt16 id);

	/**
		Deletes a record given by its uid
		@param uid the unique id of the record to be deleted
	*/
	virtual void deleteRecord(UInt32 uid);
	
};

template <class recordStructure, class packedRecordStructure> UInt32 Database<recordStructure, packedRecordStructure>::
	getRecordUid(UInt16 id)
{
	UInt32 acc_uid;
	
	DmRecordInfo(ref,id,NULL,&acc_uid,NULL);
	return acc_uid;
}

template <class recordStructure, class packedRecordStructure> UInt16 Database<recordStructure, packedRecordStructure>::
	getRecordId(UInt32 uid)
{	
	UInt16	id;
	
	DmFindRecordByID(ref,uid,&id);
	return id;
}

template <class recordStructure, class packedRecordStructure> UInt16 Database<recordStructure, packedRecordStructure>::
	getSelectedRecord() 
{ 
	if(!isRecordSelected()) 
		ErrFatalDisplay("There is no selected record");  
	return selected_record; 
}	

template <class recordStructure, class packedRecordStructure> UInt32 Database<recordStructure, packedRecordStructure>::
	getSelectedUid()
{ 
	if(!isRecordSelected()) 
		ErrFatalDisplay("There is no selected record"); 
	return getRecordUid(selected_record); 
}

template <class recordStructure, class packedRecordStructure> UInt16 Database<recordStructure, packedRecordStructure>::
	getRecordCategory(UInt16 id)
{
	UInt16	attr;

	DmRecordInfo(ref,id,&attr,NULL,NULL);
	attr &= dmRecAttrCategoryMask;	
	return attr;
}

template <class recordStructure, class packedRecordStructure> UInt16 Database<recordStructure, packedRecordStructure>::
		getRecordCategory(UInt32 uid)
{
	return getRecordCategory(getRecordId(uid));
}

template <class recordStructure, class packedRecordStructure> void Database<recordStructure, packedRecordStructure>::
		setRecordCategory(UInt16 id, UInt16 cat)
{
	UInt16	attr;

	DmRecordInfo(ref,id,&attr,NULL,NULL);
	attr &= ~dmRecAttrCategoryMask;	
	attr|=cat;
	DmSetRecordInfo(ref,id,&attr,NULL);			
}	

template <class recordStructure, class packedRecordStructure> void Database<recordStructure, packedRecordStructure>::
	setRecordCategory(UInt32 uid, UInt16 cat)
{
	setRecordCategory(getRecordId(uid),cat);
}
template <class recordStructure, class packedRecordStructure> UInt16 Database<recordStructure, packedRecordStructure>::
	findCategory(Char* cat_name)
{
	return CategoryFind(ref,cat_name);
}

template <class recordStructure, class packedRecordStructure> void Database<recordStructure, packedRecordStructure>::
	getCategoryName(UInt16 cat_id, Char* cat_name)
{
	CategoryGetName(ref, cat_id, cat_name);
}


template <class recordStructure, class packedRecordStructure> MemHandle Database<recordStructure, packedRecordStructure>::
		getUnpackedRecord(UInt32 uid,recordStructure* rec)
{
	return getUnpackedRecord(getRecordId(uid),rec);
}

template <class recordStructure, class packedRecordStructure> MemHandle Database<recordStructure, packedRecordStructure>::
		getUnpackedRecord(UInt16 id,recordStructure* rec)
{
	MemHandle				rec_h;
	packedRecordStructure*	pac_rec;

	
	rec_h=DmQueryRecord(ref,id);
	ErrFatalDisplayIf(!rec_h, "Could not find record with given ID");

	pac_rec = (packedRecordStructure*) MemHandleLock(rec_h);
	ErrFatalDisplayIf(!pac_rec, "Could not get lock on unpacked record");

	unpackRecord(rec,pac_rec);
	MemHandleUnlock(rec_h);
	return rec_h;
}
template <class recordStructure, class packedRecordStructure> MemHandle Database<recordStructure, packedRecordStructure>::
		getUnpackedRecord(recordStructure* rec)
{
	if(!isRecordSelected()) ErrFatalDisplay("There is no selected record");
	return getUnpackedRecord(selected_record, rec); 
}

template <class recordStructure, class packedRecordStructure> Database<recordStructure, packedRecordStructure>::
		Database(DmOpenRef db_open_ref)
{
	current_cat=dmAllCategories;
	ref=db_open_ref;
	unselectRecord();
}

template <class recordStructure, class packedRecordStructure> Database<recordStructure, packedRecordStructure>::
		Database(UInt32 type, UInt32 creator, Char* db_name)
{
	UInt32	mode = dmModeReadWrite;
//	Err	err = 0;
	MemPtr	c;

	created=false;
	current_cat=dmAllCategories;
	
//	appInfoType* 	appInfoP;

	// Find the database.  If it doesn't exist, create it.
	openOrCreate(type,creator, mode,0, db_name);

	//have to initialise the application info
	c = getLockedAppInfo();
	MemPtrUnlock(c);
	unselectRecord();
}

template <class recordStructure, class packedRecordStructure> Database<recordStructure, packedRecordStructure>::
		Database(UInt32 type, UInt32 creator, Char* db_name, UInt16 categ_string)
{
	UInt32	mode = dmModeReadWrite;
	MemPtr	c;

	current_cat=dmAllCategories;
	created=false;
	
	// Find the database.  If it doesn't exist, create it.
	openOrCreate(type,creator, mode,0, db_name);

	//have to initialise the application info
	c = getLockedAppInfo(categ_string);
	MemPtrUnlock(c);	
	unselectRecord();
}

template <class recordStructure, class packedRecordStructure> Database<recordStructure, packedRecordStructure>::
		~Database()
{
	DmCloseDatabase(ref);
}

template <class recordStructure, class packedRecordStructure> UInt16	Database<recordStructure, packedRecordStructure>::
		newRecord(recordStructure* rec)
{
	MemHandle	h,p;
	UInt16				index=dmMaxRecordIndex;

	//set size to 1 - will be resized by packAccount
	h=DmNewHandle(ref,1);
	ErrFatalDisplayIf(!h,"Creating new account failed. No space left on the device?");

	packRecord(rec,h);

	p = (MemHandle) MemHandleLock(h);
	MemPtrUnlock(p);
	
	DmAttachRecord(ref,&index,h,0); //automatically sets the dirty bit	

	if(current_cat!=dmAllCategories) setRecordCategory(index, current_cat);
	return index;	
}

template <class recordStructure, class packedRecordStructure> void	Database<recordStructure, packedRecordStructure>::
		deleteRecord(UInt16	id)
{
	DmRemoveRecord(ref,id);
	//if the current record was removed, unselect the record
	if(selected_record == id) unselectRecord();
}

template <class recordStructure, class packedRecordStructure> void	Database<recordStructure, packedRecordStructure>::
		deleteRecord(UInt32	uid)
{
	deleteRecord(getRecordId(uid));
}

template <class recordStructure, class packedRecordStructure> Err	Database<recordStructure, packedRecordStructure>::
		openOrCreate(UInt32 type, UInt32 creator, UInt16 mode, UInt16 cardNo, Char* name)
{
	Err	err=0;
	
	created = false;
	ErrFatalDisplayIf(cardNo,"cardno not 0");
	if (! DmFindDatabase(cardNo,name) ) { //database does not exist
		err = DmCreateDatabase(cardNo, name, creator, type, false);
		ErrFatalDisplayIf(err,"create db failed db");
		if (err) { 
			return err;
		}
		created = true; //set flag
	}

	//open database (has to exist at this point
	ref = DmOpenDatabase(cardNo,DmFindDatabase(cardNo,name),mode);

	if (! ref) return DmGetLastErr();
	
	return err;
}

template <class recordStructure, class packedRecordStructure> MemPtr	Database<recordStructure, packedRecordStructure>::
		appInfoInit(LocalID dbID,UInt16 cardNo)
{
	UInt16			wordValue;
	LocalID			appInfoID;
	appInfoType*	nilP = 0;
	appInfoType* 	appInfoP;

	MemHandle h;

	
	h = DmNewHandle(ref, sizeof (appInfoType));
	ErrFatalDisplayIf (!h,  "Cannot allocate app info block");

	//get the local id of the newly allocated appinfo record
	appInfoID = MemHandleToLocalID (h);
	//record the id
	DmSetDatabaseInfo(cardNo, dbID, NULL, NULL, NULL, NULL, NULL, NULL,NULL, &appInfoID, NULL, NULL, NULL);
	
	appInfoP = (appInfoType*) MemLocalIDToLockedPtr(appInfoID, cardNo);

	// Clear the app info block.
	DmSet (appInfoP, 0, sizeof(appInfoType), 0);

	// I don't know what this field is used for.
	wordValue = 0xFFFF;
	DmWrite (appInfoP, (UInt32)&nilP->dirtyAppInfo, &wordValue,
		sizeof(appInfoP->dirtyAppInfo));

	// Initialize the sort order. -- should be in Accountsdb
	//DmSet (appInfoP, (UInt32)&nilP->sortOrder, sizeof(appInfoP->sortOrder), soName);

	return appInfoP;
 }

template <class recordStructure, class packedRecordStructure> MemPtr	Database<recordStructure, packedRecordStructure>::
		appInfoInit(LocalID dbID,UInt16 cardNo, UInt16 categ_string)
{
	UInt16			wordValue;
	LocalID			appInfoID;
	appInfoType*	nilP = 0;
	appInfoType* 	appInfoP;

	MemHandle h;

	
	h = DmNewHandle(ref, sizeof (appInfoType));
	ErrFatalDisplayIf (!h,  "Cannot allocate app info block");

	//get the local id of the newly allocated appinfo record
	appInfoID = MemHandleToLocalID (h);
	//record the id
	DmSetDatabaseInfo(cardNo, dbID, NULL, NULL, NULL, NULL, NULL, NULL,NULL, &appInfoID, NULL, NULL, NULL);
	
	appInfoP = (appInfoType*) MemLocalIDToLockedPtr(appInfoID, cardNo);

	// Clear the app info block.
	DmSet (appInfoP, 0, sizeof(appInfoType), 0);

	// Initialize the categories.
	CategoryInitialize ((AppInfoPtr) appInfoP, categ_string);

	// I don't know what this field is used for.
	wordValue = 0xFFFF;
	DmWrite (appInfoP, (UInt32)&nilP->dirtyAppInfo, &wordValue,
		sizeof(appInfoP->dirtyAppInfo));

	// Initialize the sort order.
//	DmSet (appInfoP, (UInt32)&nilP->sortOrder, sizeof(appInfoP->sortOrder),		 soName);

	return appInfoP;
 }

template <class recordStructure, class packedRecordStructure> MemPtr	Database<recordStructure, packedRecordStructure>::
		getLockedAppInfo()
{
	UInt16	cardNo;
	LocalID	dbID;
	LocalID	appInfoID;
	Err	err;

	err = DmOpenDatabaseInfo(ref, &dbID, NULL, NULL,&cardNo, NULL);
	ErrFatalDisplayIf (err,  "Get getting application info block");
	err = DmDatabaseInfo(cardNo, dbID, NULL, NULL, NULL, NULL, NULL,NULL, NULL, &appInfoID, NULL, NULL, NULL);
	ErrFatalDisplayIf (err,  "Get getting application info block");
	
	if(appInfoID==NULL) return appInfoInit(dbID,cardNo);
	//a bit useless - to be removed unless code will be inserted here
	return MemLocalIDToLockedPtr(appInfoID, cardNo);
}

template <class recordStructure, class packedRecordStructure> void		Database<recordStructure, packedRecordStructure>::
		selectRecord(UInt32 uid) 
{ 
	selectRecord(getRecordId(uid)); 
}

template <class recordStructure, class packedRecordStructure> void		Database<recordStructure, packedRecordStructure>::
	selectRecord(UInt16 id)
{ 
	selected_record=id;
	record_selected=true; 
}

template <class recordStructure, class packedRecordStructure> MemPtr	Database<recordStructure, packedRecordStructure>::
		getLockedAppInfo(UInt16 categ_string)
{
	UInt16	cardNo;
	LocalID	dbID;
	LocalID	appInfoID;
	Err	err;

	err = DmOpenDatabaseInfo(ref, &dbID, NULL, NULL,&cardNo, NULL);
	ErrFatalDisplayIf (err,  "Get getting application info block");
	err = DmDatabaseInfo(cardNo, dbID, NULL, NULL, NULL, NULL, NULL,NULL, NULL, &appInfoID, NULL, NULL, NULL);
	ErrFatalDisplayIf (err,  "Get getting application info block");
	
	if(appInfoID==NULL) return appInfoInit(dbID, cardNo, categ_string);
	//a bit useless - to be removed unless code will be inserted here
	return MemLocalIDToLockedPtr(appInfoID, cardNo);
}

#endif
