/**	
	Code to access the database containing the Transaction information

	$Revision: 1.29 $
	Last updated: $Date: 2001/07/23 23:36:04 $

	Copyright (C) 2000-2001 Vlad Mereuta <dizzy@users.sourceforge.net>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
	
	$Log: TransactionsDB.h,v $
	Revision 1.29  2001/07/23 23:36:04  chaese
	Write scheduling info to database for scheduled transactions only
	
	Revision 1.28  2001/07/23 06:22:32  dizzy
	Scheduling bugfixes, Some modifications to the TransactionStructure
	
	Revision 1.27  2001/07/15 11:38:12  dizzy
	Added interface for purging transactions
	
	Revision 1.26  2001/07/09 02:29:39  chaese
	implement splits GUI & other small improvements along the way. The bulk of
	the splits GUI is complete, small data integrity checks are coming soon
	(need to prevent multiple splits for the same account within one transaction
	and stuff like that...)
	
	Revision 1.25  2001/07/02 14:03:54  dizzy
	Horde of bugfixes to the transaction editing routines...
	
	Revision 1.24  2001/07/02 11:03:04  dizzy
	Cleaned up TransactionsForm.
	Removed details button (tap will take user to details)
	Unbalanced transactions are marked
	The total in the Scheduled view is the total of all scheduled transactions (as opposed to the account balance)
	
	Revision 1.23  2001/07/01 12:21:55  dizzy
	Merged ScheduledDB into TransactionsDB
	
	Revision 1.22  2001/06/27 22:58:11  dizzy
	Restructured #include directives
	
	Revision 1.21  2001/06/27 01:25:18  dizzy
	added purge transaction function
	
	Revision 1.20  2001/06/23 04:51:48  chaese
	fixed bug #435294 and minor cleanups
	
	Revision 1.19  2001/06/11 11:03:21  chaese
	Increased display speed in account and transaction tables:
	* Associate table rows with the record index rather than UID. (Minor gain)
	* Perform preference lookup at program startup rather than every time a
	date or number is drawn. (Major gain! That was really stupid of me to put
	the lookups into the drawing function in the first place!)
	
	Revision 1.18  2001/06/10 04:05:41  chaese
	basic infrastructure for split transactions (somewhat tested, but not
	guaranteed to be bug free)
	CAUTION! Database format has changed! Expect nastiness when you run the
	new version on an old Database.
	
	Revision 1.17  2001/04/21 12:54:59  dizzy
	Added number field for transactions
	Added credit checkbox (the payments/withdrawals are display with inverted sign)
	
	Revision 1.16  2001/01/29 18:49:30  dizzy
	Put code for scheduled transactions in a separate code section to enable the linking process to succeed.
	Scheduled trans info can be add/deleted/updated
	Added cleared field to trans structure
	
	Revision 1.15  2001/01/23 20:40:11  dizzy
	Initial support for scheduled transactions.
	
	Revision 1.14  2000/12/13 16:37:34  dizzy
	Ability to display a column which indicates/modifies the cleared state for each transaction
		(it is on by default; it can be turned off from the preferences dialog)
	The default view in the transaction list can be selected from the preferences dialog
	
	Revision 1.13  2000/12/11 16:27:25  dizzy
	C++ version of the code.	
*/

#ifndef __TRANSACTIONS_DB__
#define __TRANSACTIONS_DB__

#include "Database.h"


/// type of the Transaction database
#define TRANSACTIONSDB_TYPE	'DATA'
/// db name
#define TRANSACTIONSDB_NAME	"FreeCoins-Transactions"

/// payment method
typedef enum { pm_card=0, pm_cash=1, pm_cheque=2, pm_so=3, pm_dd=4,
	 pm_transfer=5 } payment_method;
	
typedef enum { rep_day, rep_week, rep_month, rep_year } repeatInterval;


///unpacked structure of an scheduled scheduled
typedef struct {
   ///interval at which the transaction will repeat (day,week,etc.)
	repeatInterval    repeat_interval;
	/// day of the week/month on which transaction will repeat
	UInt16              repeat_on;
	///frequency of the repetition (i.e. every n days)
	UInt16              frequency;
	///number of times the transaction will be repeated (-1 - forever)
	Int16                repeat_times;
	///the last date at which the transaction can be repeated (start date is stored in the transaction itself)
	DateType		end_date;
	///if true, end_date defines the end of this repeat transaction; otherwise repeat_times defines it.
	Boolean           end_by_date;
} scheduledStructure;

///the packed structure of a transaction - see the unpacked struct for explanations
typedef struct {
	DateType date;
	UInt8 num_splits;
	UInt8 method;
	unsigned is_scheduled	:1;
	unsigned reserved		:15;
} packedTransactionStructure;

/// data specific to scheduled transactions
typedef struct {
	UInt16 repeat_on;
	UInt16 frequency;
	Int16 repeat_times;
	DateType end_date;
    unsigned to_be_cleared	:1;
    unsigned end_by_date	:1;
    unsigned reserved		:14;
	repeatInterval repeat_interval;
} packedScheduleStructure;

/**
   unpacked structure of an Transaction
   \note the fields related to scheduling should only be considered if the transaction itself is marked as 
   schduled (i.e. is in the scheduled category)
*/
typedef struct {
	///date on which the transaction was made
	DateType		date;
	///the method of payment
	payment_method	method;
	/// indicates if the transaction is scheduled
	Boolean is_scheduled;
	///this flag indicates whether the record is to be cleared (for scheduled transactions)
	Boolean			to_be_cleared;
	/// number of splits
	UInt8 			num_splits;
	///contains scheduling information
	scheduledStructure sched;
	/// person/company 
	Char*			payee;
	///optional number to id this transaction (cheque number, order number, etc.)
	Char*			num;
	/// note on the transaction	
	Char*			note;
} transactionStructure;

/**
   class to manipulate the transactions database.
   \note Transactions can belong to one of the three categories:
      -# cleared
      -# uncleared
      -# scheduled
   Cleared/uncleared transaction should be treated (\em mostly)  the same. The scheduled transactions
   contain the information for the transaction <b> to be scheduled </b>in the transaction body, and information
   about how this is to be done in the scheduled structure
*/
class TransactionsDB : public Database <transactionStructure, packedTransactionStructure>
{
 public:
	/// unpacks an Transaction record
	void	unpackRecord (transactionStructure* trans, packedTransactionStructure* packed_trans);

	/**
       packs an Transaction record
       \todo we should be able to pack the scheduled structure to 0 or 1 bytes. I am not sure how 
       viable this is (as we do not always know whether the transaction is scheduled or not when packing it ) but if it
       can be done it would save about 14 bytes for each transaction
   */
	void	packRecord (transactionStructure* trans, MemHandle db_entry);

	/** 
		get the date of the transaction
		@param id record index
		@param pDate pointer to allocated date structure that will receive the date
	*/
	void	getDate(UInt16 id, DateType * pDate);

	/**
		determine if the transaction is balanced
		@param trans_uid UID of the transaction
		@return true if the transaction is balanced, false otherwise
	*/
	Boolean isBalanced(UInt32 trans_uid);

	/**
		get the amount by which the transaction affects the given account
		@param trans_uid UID of the transaction
		@param acc_uid UID of the account
		@return signed amount
	*/
	Int32 getAmount(UInt32 trans_uid, UInt32 acc_uid);

	/**
		get the peer account in the given transaction from the point of view
		of the given account. Only makes sense for transactions that are
		split two ways, i.e. simple transfers.
		@param trans_uid UID of the transaction
		@param acc_uid UID of the account
		@return UID of the peer account
	*/
	UInt32 getPeerAcc(UInt32 trans_uid, UInt32 acc_uid);

	/**
		checks whether a record is cleared or not
		@param id id of the record to be checked
		@return true if the record is cleared, false otherwise
	*/
	Boolean	isCleared(UInt16 id);

	/**
		checks whether a record is to be cleared or not
		@param id id of the record to be checked
		@return true if the record is set to be cleared in the future, false otherwise
	*/
	Boolean	willBeCleared(UInt16 id);

   /**
      checks whether a transaction is scheduled or not by checking the record category
      @param id the id of the transaction to be checked
      @return true if it is scheduled, false othewise
   */
   Boolean	isScheduled(UInt16 id);

	/**
		set a record category to cleared
		@param id id of the record to be set
	*/
	void	clearRecord(UInt16 id);

	/**
		set a record category to uncleared
		@param id id of the record to be set
	*/
	void	unclearRecord(UInt16 id);

	/// initialize the database
	TransactionsDB();

	///init the Transactions DB with some default records
	void	initializeTransactions();

	/**
		creates a new record and update the balances of the two involved accounts
		@param s pointer to the structure containing the transaction data
		@param cleared boolean which indicates the status of the record. The function will place the record in the right category
		@param scheduled (which defaults to false if no value is given) indicates whether the transaction is scheduled or not. the 
			cateogry will be set to the right one
		@return id of the newly created record
	*/
	UInt16	newRecord(transactionStructure* s, Boolean cleared, Boolean scheduled = false);
	
	/**
		deletes a record, given its id
		@param id the id of the record to be deleted
		@param keep_balance if true then the balances of the accounts involved in the transaction will *not* be changed
	*/
	void deleteRecord(UInt16 id, Boolean keep_balance = false);

	/**
		Compares two packed transaction structures.
		For parameters see the PalmOS reference
	*/
	static Int16	compareRecords(packedTransactionStructure* rec1,packedTransactionStructure* rec2,Int16 other,SortRecordInfoPtr sr1,SortRecordInfoPtr sr2,MemHandle app_info);

	/** 
		This function deletes all the transactions earlier than the given date. 
		This should not change the existing balances for the accounts involved. Both cleared and uncleared transactions are removed, but
		the scheduled ones are kept
   
   	\todo message informing the operation is being performed while this function runs
      \warning should probably set the archived bit on deletion rather than removing the records. this is true for all the destructive operations, but
      there is really no point in doing so before the conduit reaches a working state      
      \note scheduled transaction are not deleted on purging
		@param last_date the last date up to which the transactions are to be deleted 
	*/
  void purgeTransactions(DateType last_date);
  
	/**
		records the next date in the associated transaction (in relation to the date already recorded)
		if we have repeated the record enought times, the record is deleted
		@param id id of the record for which to find the next date
		@return true if the record has been deleted, false otherwise
	*/
	Boolean calculateNextDate(UInt16 id);
	
	/**
		scans through all the scheduled transactions, and if there are any due they are recorded and their next occurance date
		is updated
		@param force if true then the current date is ignored the update is perfomed anyway
		@return true if updates were made, false otherwise
	*/
	Boolean schedule(Boolean force=false);  

   	Boolean scheduleNextRecord();  
   
   /** 
      returns the total for the scheduled transactions in a given account
      @param account_uid uid of the account to be checked
      @return the sum of the amounts of all scheduled transaction for that account
   */
   Int32      getScheduledTotal(UInt32 account_uid);
};

extern TransactionsDB* transactions_db;

#endif
